"""

Functions for reading and writing data.

"""


from os import listdir as _listdir, path as _path
import re as _re
import json as _json
from .fixation import FixationSequence as _FixationSequence
from .text import TextBlock as _TextBlock


def read(file_path):
    """

    Read in a JSON file. `eyekit.fixation.FixationSequence` and
    `eyekit.text.TextBlock` objects are automatically decoded and
    instantiated.

    """
    with open(file_path, encoding="utf-8") as file:
        data = _json.load(file, object_hook=_eyekit_decoder)
    return data


def write(data, file_path, compress=True):
    """

    Write arbitrary data to a JSON file. If `compress` is `True`, the file is
    written in the most compact way; if `False`, the file will be larger but
    more human-readable. `eyekit.fixation.FixationSequence` and
    `eyekit.text.TextBlock` objects are automatically serialized.

    """
    if compress:
        indent = None
        separators = (",", ":")
    else:
        indent = "\t"
        separators = (", ", " : ")
    with open(file_path, "w", encoding="utf-8") as file:
        _json.dump(
            data,
            file,
            default=_eyekit_encoder,
            ensure_ascii=False,
            indent=indent,
            separators=separators,
        )


def import_asc(file_path, trial_begin_var, trial_begin_vals, extract_vars=[]):
    """

    Import a single ASC file or a directory of ASC files. The importer looks
    for a `trial_begin_var` that is set to one of the `trial_begin_vals`, and
    then extracts all `EFIX` lines that occur within the subsequent
    `START`–`END` block. Optionally, you can specify other variables that you
    want to extract, resulting in imported data that looks like this:

    ```
    {
      "trial_0" : {
        "trial_type" : "Experimental",
        "passage_id" : "passage_a",
        "response" : "yes",
        "fixations" : FixationSequence[[368, 161, 208], ..., [562, 924, 115]]
      }
    }
    ```

    """
    if _path.isfile(file_path):
        file_paths = [file_path]
    else:
        file_paths = [
            _path.join(file_path, filename)
            for filename in _listdir(file_path)
            if filename.endswith(".asc")
        ]
    if isinstance(trial_begin_vals, str):
        trial_begin_vals = [trial_begin_vals]
    trial_line_regex = _re.compile(
        r"^.+?TRIAL_VAR\s(?P<var>("
        + "|".join([trial_begin_var] + extract_vars)
        + r")?)\s(?P<val>.+?)$"
    )
    efix_line_regex = _re.compile(
        r"^EFIX R\s+(?P<stime>.+?)\s+(?P<etime>.+?)\s+(?P<duration>.+?)\s+(?P<x>.+?)\s+(?P<y>.+?)\s"
    )
    data = {}
    for file_path in file_paths:
        curr_trial = {}
        start_flag = False
        with open(file_path) as file:
            for line in file:
                if start_flag:
                    if line.startswith("EFIX"):
                        line_match = efix_line_regex.match(line)
                        if line_match:
                            curr_trial["fixations"].append(
                                (
                                    int(round(float(line_match["x"]), 0)),
                                    int(round(float(line_match["y"]), 0)),
                                    int(line_match["duration"]),
                                )
                            )
                    elif line.startswith("END"):
                        curr_trial["fixations"] = _FixationSequence(
                            curr_trial["fixations"]
                        )
                        data[f"trial_{len(data)}"] = curr_trial
                        start_flag = False
                        curr_trial = {}
                else:
                    if line.startswith("MSG"):
                        line_match = trial_line_regex.match(line)
                        if line_match:
                            if (
                                line_match["var"] == trial_begin_var
                                and line_match["val"] not in trial_begin_vals
                            ):
                                continue
                            curr_trial[line_match["var"]] = line_match["val"]
                    elif line.startswith("START") and trial_begin_var in curr_trial:
                        start_flag = True
                        curr_trial["fixations"] = []
    return data


def _eyekit_encoder(obj):
    """

    Convert a `FixationSequence` or `TextBlock` object into something JSON
    serializable that can later be decoded by _eyekit_decoder().

    """
    if isinstance(obj, _FixationSequence):
        return {"__FixationSequence__": obj._serialize()}
    if isinstance(obj, _TextBlock):
        return {"__TextBlock__": obj._serialize()}
    raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")


def _eyekit_decoder(obj):
    """

    Decode an object into a `FixationSequence` or `TextBlock` if the key
    implies that it is one of those types.

    """
    if "__FixationSequence__" in obj:
        return _FixationSequence(obj["__FixationSequence__"])
    if "__TextBlock__" in obj:
        return _TextBlock(**obj["__TextBlock__"])
    return obj



"""

Defines the `Fixation` and `FixationSequence` objects, which are used to
represent fixation data.

"""


import numpy as _np


class Fixation:

    """

    Representation of a single fixation event. It is not usually necessary to
    create `Fixation` objects manually; they are created automatically during
    the instantiation of a `FixationSequence`.

    """

    def __init__(self, x: int, y: int, duration: int, discarded: bool = False):
        self.x = x
        self.y = y
        self.duration = duration
        self.discarded = discarded

    def __repr__(self):
        return f"Fixation[{self.x},{self.y}]"

    @property
    def x(self) -> int:
        """X-coordinate of the fixation."""
        return self._x

    @x.setter
    def x(self, x):
        self._x = int(x)

    @property
    def y(self) -> int:
        """Y-coordinate of the fixation."""
        return self._y

    @y.setter
    def y(self, y):
        self._y = int(y)

    @property
    def xy(self) -> tuple:
        """XY-coordinates of the fixation."""
        return self._x, self._y

    @xy.setter
    def xy(self, xy):
        self._x = int(xy[0])
        self._y = int(xy[1])

    @property
    def duration(self) -> int:
        """Duration of the fixation in milliseconds."""
        return self._duration

    @duration.setter
    def duration(self, duration):
        self._duration = int(duration)

    @property
    def discarded(self) -> bool:
        """`True` if the fixation has been discarded, `False` otherwise."""
        return self._discarded

    @discarded.setter
    def discarded(self, discarded):
        self._discarded = bool(discarded)

    @property
    def tuple(self) -> tuple:
        """Tuple representation of the fixation."""
        if self.discarded:
            return (self._x, self._y, self._duration, True)
        return (self._x, self._y, self._duration)


class FixationSequence:

    """

    Representation of a sequence of consecutive fixations, typically from a
    single trial.

    """

    def __init__(self, sequence: list = []):
        """Initialized with:

        - `sequence` List of tuples of ints, or something similar, that
        conforms to the following structure: `[(106, 540, 100), (190, 536,
        100), ..., (763, 529, 100)]`, where each tuple contains the
        X-coordinate, Y-coordinate, and duration of a fixation

        """
        self._sequence = []
        for fixation in sequence:
            self.append(fixation)

    def __repr__(self):
        if len(self) > 2:
            return f"FixationSequence[{str(self._sequence[0])}, ..., {str(self._sequence[-1])}]"
        elif len(self) == 2:
            return (
                f"FixationSequence[{str(self._sequence[0])}, {str(self._sequence[1])}]"
            )
        elif len(self) == 1:
            return f"FixationSequence[{str(self._sequence[0])}]"
        return "FixationSequence[]"

    def __len__(self):
        return len(self._sequence)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._sequence[index]
        if isinstance(index, slice):
            return FixationSequence(self._sequence[index.start : index.stop])
        raise IndexError("Index to FixationSequence must be integer or slice")

    def __iter__(self):
        for fixation in self._sequence:
            yield fixation

    def __add__(self, other):
        if not isinstance(other, FixationSequence):
            raise TypeError("Can only concatenate with another FixationSequence")
        return FixationSequence(self._sequence + other._sequence)

    def append(self, fixation):
        if not isinstance(fixation, Fixation):
            try:
                fixation = Fixation(*fixation)
            except:
                raise ValueError(
                    "Cannot create FixationSequence, pass a list of (x, y, duration) for each fixation"
                )
        self._sequence.append(fixation)

    def copy(self, include_discards=True):
        """

        Returns a copy of the fixation sequence.

        """
        if include_discards:
            return FixationSequence(
                [fixation.tuple for fixation in self.iter_with_discards()]
            )
        return FixationSequence(
            [fixation.tuple for fixation in self.iter_without_discards()]
        )

    def purge(self):
        """

        Permanently removes all discarded fixations from the fixation sequence.

        """
        self._sequence = [
            fixation for fixation in self._sequence if not fixation.discarded
        ]

    def iter_with_discards(self):
        """

        Iterates over the fixation sequence including any discarded fixations.
        This is also the default behavior when iterating over a
        `FixationSequence` directly.

        """
        for fixation in self._sequence:
            yield fixation

    def iter_without_discards(self):
        """

        Iterates over the fixation sequence without any discarded fixations.

        """
        for fixation in self._sequence:
            if not fixation.discarded:
                yield fixation

    def XYarray(self, include_discards=False):
        """

        Returns a Numpy array containing the XY-coordinates of the fixations.

        """
        if include_discards:
            return _np.array(
                [fixation.xy for fixation in self.iter_with_discards()], dtype=int
            )
        return _np.array(
            [fixation.xy for fixation in self.iter_without_discards()], dtype=int
        )

    def Xarray(self, include_discards=False):
        """

        Returns a Numpy array containing the X-coordinates of the fixations.

        """
        if include_discards:
            return _np.array(
                [fixation.x for fixation in self.iter_with_discards()], dtype=int
            )
        return _np.array(
            [fixation.x for fixation in self.iter_without_discards()], dtype=int
        )

    def Yarray(self, include_discards=False):
        """

        Returns a Numpy array containing the Y-coordinates of the fixations.

        """
        if include_discards:
            return _np.array(
                [fixation.y for fixation in self.iter_with_discards()], dtype=int
            )
        return _np.array(
            [fixation.y for fixation in self.iter_without_discards()], dtype=int
        )

    def _serialize(self):
        """

        Returns representation of the fixation sequence in simple list format
        for serialization.

        """
        return [fixation.tuple for fixation in self.iter_with_discards()]



import re
import cairocffi as cairo

regex_italic = re.compile(" italic", re.IGNORECASE)
regex_bold = re.compile(" bold", re.IGNORECASE)


class Font:

    """

    Wrapper around Cairo's font selection mechanism.

    """

    def __init__(self, face, size):
        if regex_italic.search(face):
            self.slant = "italic"
            slant = cairo.FONT_SLANT_ITALIC
            face = regex_italic.sub("", face)
        else:
            self.slant = "normal"
            slant = cairo.FONT_SLANT_NORMAL
        if regex_bold.search(face):
            self.weight = "bold"
            weight = cairo.FONT_WEIGHT_BOLD
            face = regex_bold.sub("", face)
        else:
            self.weight = "normal"
            weight = cairo.FONT_WEIGHT_NORMAL
        self.family = face.strip()
        self.size = float(size)

        self.toy_font_face = cairo.ToyFontFace(self.family, slant, weight)
        self.scaled_font = cairo.ScaledFont(
            self.toy_font_face, cairo.Matrix(xx=self.size, yy=self.size)
        )

    def calculate_width(self, text):
        return self.scaled_font.text_extents(text)[4]

    def calculate_height(self, text):
        return self.scaled_font.text_extents(text)[3]



"""

Functions for performing common procedures, such as discarding out of
bounds fixations and snapping fixations to the lines of text.

"""


from os.path import splitext as _splitext
import cairocffi as _cairo
from ._core import distance as _distance
from .fixation import FixationSequence as _FixationSequence
from .text import TextBlock as _TextBlock
from . import _drift


def discard_short_fixations(fixation_sequence, threshold=50):
    """

    Given a `eyekit.fixation.FixationSequence`, discard all fixations that are
    shorter than some threshold value. Operates directly on the sequence and
    does not return a copy.

    """
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type eyekit.FixationSequence")
    for fixation in fixation_sequence:
        if fixation.duration < threshold:
            fixation.discarded = True


def discard_out_of_bounds_fixations(fixation_sequence, text_block, threshold=100):
    """

    Given a `eyekit.fixation.FixationSequence` and `eyekit.text.TextBlock`,
    discard all fixations that do not fall within some threshold distance of
    any character in the text. Operates directly on the sequence and does not
    return a copy.

    """
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type eyekit.FixationSequence")
    if not isinstance(text_block, _TextBlock):
        raise TypeError("text_block should be of type eyekit.TextBlock")
    for fixation in fixation_sequence:
        for char in text_block:
            if _distance(fixation.xy, char.center) < threshold:
                break
        else:  # For loop exited normally, so no char was within the threshold
            fixation.discarded = True


def snap_to_lines(fixation_sequence, text_block, method="warp", **kwargs):
    """

    Given a `eyekit.fixation.FixationSequence` and `eyekit.text.TextBlock`,
    snap each fixation to the line that it most likely belongs to, eliminating
    any y-axis variation or drift. Operates directly on the sequence and does
    not return a copy. Several methods are available, some of which take
    optional parameters or require SciPy to be installed. For a full
    description and evaluation of these methods, see [Carr et al.
    (2020)](https://osf.io/jg3nc/).

    """
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type eyekit.FixationSequence")
    if not isinstance(text_block, _TextBlock):
        raise TypeError("text_block should be of type eyekit.TextBlock")
    if method not in _drift.methods:
        raise ValueError(
            f"Invalid method. Supported methods are: {', '.join(_drift.methods)}"
        )
    if text_block.n_rows == 1:
        for fixation in fixation_sequence.iter_without_discards():
            fixation.y = text_block.line_positions[0]
    else:
        fixation_XY = fixation_sequence.XYarray(include_discards=False)
        corrected_Y = _drift.methods[method](fixation_XY, text_block, **kwargs)
        for fixation, y in zip(fixation_sequence.iter_without_discards(), corrected_Y):
            fixation.y = y


# Append the docstring from each of the methods
snap_to_lines.__doc__ += "\n\n" + "\n\n".join(
    [
        f"- `{method}` : " + func.__doc__.strip()
        for method, func in _drift.methods.items()
    ]
)


def align_to_screenshot(
    text_block,
    screenshot_path,
    output_path=None,
    show_text=True,
    show_bounding_boxes=False,
):
    """

    Create an image depicting a PNG screenshot overlaid with a
    `eyekit.text.TextBlock` in green. If no output path is provided, the
    output is written to the same directory as the screenshot file. This is
    useful for establishing the correct `eyekit.text.TextBlock` parameters to
    match what participants are actually seeing.

    """
    surface = _cairo.ImageSurface(_cairo.FORMAT_ARGB32, 1, 1).create_from_png(
        screenshot_path
    )
    context = _cairo.Context(surface)
    screen_width = surface.get_width()
    screen_height = surface.get_height()
    context.set_source_rgb(0.60392, 0.80392, 0.19607)
    context.set_font_face(text_block._font.toy_font_face)
    context.set_font_size(text_block._font.size)
    context.set_line_width(2)
    context.move_to(text_block.x_tl, 0)
    context.line_to(text_block.x_tl, screen_height)
    context.stroke()
    for i, line in enumerate(text_block.lines()):
        if i == 0:
            context.move_to(0, text_block._first_baseline)
            context.line_to(screen_width, text_block._first_baseline)
            context.stroke()
            context.set_dash([8, 4])
        else:
            context.move_to(text_block.x_tl, line.baseline)
            context.line_to(screen_width, line.baseline)
            context.stroke()
        if show_text:
            context.move_to(line.x_tl, line.baseline)
            context.show_text(line.text)
    if show_bounding_boxes:
        context.set_dash([])
        for word in text_block.words(add_padding=False):
            context.rectangle(*word.box)
            context.stroke()
    if output_path is None:
        output_path = _splitext(screenshot_path)[0] + "_eyekit.png"
    surface.write_to_png(output_path)


def font_size_at_72dpi(font_size, at_dpi=96):
    """

    Convert a font size at some dpi to the equivalent font size at 72dpi.
    Typically, this can be used to convert a Windows-style 96dpi font size to
    the equivalent size at 72dpi.

    """
    return font_size * at_dpi / 72



special_characters_by_language = {
    "Bulgarian": "АБВГДЕЖЗИЙЍКЛМНОПРСТУФХЦЧШЩЪЬЮЯабвгдежзийѝклмнопрстуфхцчшщъьюя",
    "Croatian": "ČĆĐŠŽčćđšž",
    "Czech": "ÁČĎÉĚÍŇÓŘŠŤÚŮÝŽáčďéěíňóřšťúůýž",
    "Danish": "ÆØÅæøå",
    "Estonian": "ŠŽÕÄÖÜšžõäöü",
    "Finnish": "ÅÄÖåäö",
    "French": "ÇÉÀÈÙÂÊÎÔÛËÏÜŸŒçéàèùâêîôûëïüÿœ",
    "German": "ÄÖÜäöüß",
    "Greek": "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩΆΈΉΊΌΎΏΪΫαβγδεζηθικλμνξοπρσςτυφχψωάέήίόύώϊϋΐΰ",
    "Hungarian": "ÁÉÍÓÖŐÚÜŰáéíóöőúüű",
    "Irish": "ÁÉÍÓÚáéíóú",
    "Italian": "ÀÉÈÍÌÓÒÚÙàéèíìóòúù",
    "Latvian": "ĀČĒĢĪĶĻŅŠŪŽāčēģīķļņšūž",
    "Lithuanian": "ĄČĘĖĮŠŲŪŽąčęėįšųūž",
    "Maltese": "ĊĠĦŻċġħż",
    "Norwegian": "ÆØÅæøå",
    "Polish": "ĄĆĘŁŃÓŚŹŻąćęłńóśźż",
    "Portuguese": "ÁÂÃÀÇÉÊÍÓÔÕÚáâãàçéêíóôõú",
    "Romanian": "ĂÂÎȘȚăâîșț",
    "Russian": "АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдеёжзийклмнопрстуфхцчшщъыьэюя",
    "Slovak": "ÁÄČĎÉÍĹĽŇÓÔŔŠŤÚÝŽáäčďéíĺľňóôŕšťúýž",
    "Slovene": "ČŠŽčšž",
    "Spanish": "ÑÁÉÍÓÚÜñáéíóúü",
    "Swedish": "ÅÄÖåäö",
    "Turkish": "ÇĞİÖŞÜçşğıiöü",
}

special_characters = []
for characters in special_characters_by_language.values():
    special_characters.extend(characters)

characters = "A-Za-z" + "".join(set(special_characters))



"""
.. include:: ../.GUIDE.md
   :start-line: 4
"""

from .fixation import FixationSequence
from .text import TextBlock
from . import analysis
from . import io
from . import tools
from . import vis

del fixation, text

try:
    from ._version import __version__
except ImportError:
    __version__ = "???"



"""

Defines the `TextBlock` and `InterestArea` objects for handling texts.

"""


import re as _re
import cairocffi as _cairo
from . import _alpha, _font


class Box(object):

    """

    Representation of a bounding box, which provides an underlying framework
    for `Character`, `InterestArea`, and `TextBlock`.

    """

    def __contains__(self, fixation):
        if (self.x_tl <= fixation.x <= self.x_br) and (
            self.y_tl <= fixation.y <= self.y_br
        ):
            return True
        return False

    @property
    def x(self) -> float:
        """X-coordinate of the center of the bounding box"""
        return self.x_tl + self.width / 2

    @property
    def y(self) -> float:
        """Y-coordinate of the center of the bounding box"""
        return self.y_tl + self.height / 2

    @property
    def x_tl(self) -> float:
        """X-coordinate of the top-left corner of the bounding box"""
        return self._x_tl

    @property
    def y_tl(self) -> float:
        """Y-coordinate of the top-left corner of the bounding box"""
        return self._y_tl

    @property
    def x_br(self) -> float:
        """X-coordinate of the bottom-right corner of the bounding box"""
        return self._x_br

    @property
    def y_br(self) -> float:
        """Y-coordinate of the bottom-right corner of the bounding box"""
        return self._y_br

    @property
    def width(self) -> float:
        """Width of the bounding box"""
        return self.x_br - self.x_tl

    @property
    def height(self) -> float:
        """Height of the bounding box"""
        return self.y_br - self.y_tl

    @property
    def box(self) -> tuple:
        """The bounding box represented as x_tl, y_tl, width, and height"""
        return self.x_tl, self.y_tl, self.width, self.height

    @property
    def center(self) -> tuple:
        """XY-coordinates of the center of the bounding box"""
        return self.x, self.y


class Character(Box):

    """

    Representation of a single character of text. A `Character` object is
    essentially a one-letter string that occupies a position in space and has
    a bounding box. It is not usually necessary to create `Character` objects
    manually; they are created automatically during the instantiation of a
    `TextBlock`.

    """

    def __init__(self, char, x_tl, y_tl, x_br, y_br, baseline):
        if isinstance(char, str) and len(char) == 1:
            self._char = char
        else:
            raise ValueError("char must be one-letter string")
        self._x_tl, self._y_tl = float(x_tl), float(y_tl)
        self._x_br, self._y_br = float(x_br), float(y_br)
        self._baseline = float(baseline)

    def __repr__(self):
        return self._char

    @property
    def baseline(self) -> float:
        """The y position of the character baseline"""
        return self._baseline


class InterestArea(Box):

    """

    Representation of an interest area – a portion of a `TextBlock` object
    that is of potential interest. It is not usually necessary to create
    `InterestArea` objects manually; they are created automatically when you
    slice a `TextBlock` object or when you iterate over lines, words,
    characters, ngrams, or zones parsed from the raw text.

    """

    def __init__(self, chars, id, padding=0):
        for char in chars:
            if not isinstance(char, Character):
                raise ValueError("chars must only contain Character objects")
        self._x_tl = chars[0].x_tl - padding
        self._y_tl = chars[0].y_tl
        self._x_br = chars[-1].x_br + padding
        self._y_br = chars[-1].y_br
        self._chars = chars
        self._id = str(id)

    def __repr__(self):
        return f"InterestArea[{self.id}, {self.text}]"

    def __getitem__(self, key):
        return self._chars[key]

    def __len__(self):
        return len(self._chars)

    def __iter__(self):
        for char in self._chars:
            yield char

    @property
    def id(self) -> str:
        """

        Interest area ID. By default, these ID's have the form 1:5:10, which
        represents the line number and column indices of the `InterestArea` in
        its parent `TextBlock`. However, IDs can also be changed to any
        arbitrary string.

        """
        return self._id

    @id.setter
    def id(self, id):
        self._id = str(id)

    @property
    def baseline(self) -> float:
        """The y position of the text baseline"""
        return self._chars[0].baseline

    @property
    def text(self) -> str:
        """String representation of the interest area"""
        return "".join(map(str, self._chars))


class TextBlock(Box):

    """

    Representation of a piece of text, which may be a word, sentence, or
    entire multiline passage.

    """

    _default_position = (0, 0)
    _default_font_face = "Courier New"
    _default_font_size = 20.0
    _default_line_height = None
    _default_alphabet = None

    _alpha_solo = _re.compile(f"[{_alpha.characters}]")
    _alpha_plus = _re.compile(f"[{_alpha.characters}]+")
    _zone_markup = _re.compile(r"(\[(.+?)\]\{(.+?)\})")

    @classmethod
    def defaults(
        cls,
        position: tuple = None,
        font_face: str = None,
        font_size: float = None,
        line_height: float = None,
        alphabet: str = None,
    ):
        """

        Set default `TextBlock` parameters. If you plan to create several
        `TextBlock`s with the same parameters, it may be useful to set the
        default parameters at the top of your script or at the start of your
        session:

        ```python
        import eyekit
        eyekit.TextBlock.defaults(font_face='Helvetica')
        txt = eyekit.TextBlock('The quick brown fox')
        print(txt.font_face) # 'Helvetica'
        ```

        """
        if position is not None:
            cls._default_position = (float(position[0]), float(position[1]))
        if font_face is not None:
            cls._default_font_face = str(font_face)
        if font_size is not None:
            cls._default_font_size = float(font_size)
        if line_height is not None:
            cls._default_line_height = float(line_height)
        if alphabet is not None:
            cls._default_alphabet = str(alphabet)
            cls._alpha_solo = _re.compile(f"[{cls._default_alphabet}]")
            cls._alpha_plus = _re.compile(f"[{cls._default_alphabet}]+")

    def __init__(
        self,
        text: list,
        position: tuple = None,
        font_face: str = None,
        font_size: float = None,
        line_height: float = None,
        alphabet: str = None,
    ):
        """
        Initialized with:

        - ```text``` The line of text (string) or lines of text (list of
        strings). Optionally, these can be marked up with arbitrary interest
        areas (zones); for example, ```The quick brown fox
        jump[ed]{past-suffix} over the lazy dog```.

        - `position` XY-coordinates of the left edge of the first baseline of
        the block of text.

        - `font_face` Name of a font available on your system. The keywords
        `italic` and/or `bold` can also be included to select the desired
        style, e.g., `Minion Pro bold italic`.

        - `font_size` Font size in points (at 72dpi). To convert a font size
        from some other dpi, use `eyekit.tools.font_size_at_72dpi()`.

        - `line_height` Height of a line of text in points. Generally
        speaking, for single line spacing, the line height is equal to the
        font size, for double line spacing, the line height is equal to 2 ×
        the font size, etc. By default, the line height is assumed to be the
        same as the font size (single line spacing). This parameter also
        effectively determines the height of the bounding boxes around
        interest areas.

        - `alphabet` A string of characters that are to be considered
        alphabetical, which determines, for example, what is considered a
        word. By default, Eyekit considers the standard Latin, Greek, and
        Cyrillic alphabets to be alphabetical, plus the special and accented
        characters from most European languages. However, if you need support
        for some other alphabet, or if you want to modify Eyekit's default
        behavior, you can set an alternative alphabet here. This parameter is
        case sensitive, so you must supply upper- and lower-case variants. For
        example, if you wanted to treat apostrophes and hyphens as
        alphabetical, you might use `alphabet="A-Za-z'-"`. This would allow,
        for example, "Where's the orang-utan?" to be treated as three words
        rather than five.

        """

        # TEXT
        if isinstance(text, str):
            self._text = [text]
        elif isinstance(text, list):
            self._text = [str(line) for line in text]
        else:
            raise ValueError("text should be a string or a list of strings")

        # POSITION
        if position is None:
            self._x_tl = self._default_position[0]
            self._first_baseline = self._default_position[1]
        else:
            self._x_tl = float(position[0])
            self._first_baseline = float(position[1])

        # FONT FACE
        if font_face is None:
            self._font_face = self._default_font_face
        else:
            self._font_face = str(font_face)

        # FONT SIZE
        if font_size is None:
            self._font_size = self._default_font_size
        else:
            self._font_size = float(font_size)

        # LINE HEIGHT
        if line_height is None:
            if self._default_line_height is None:
                self._line_height = self._font_size
            else:
                self._line_height = self._default_line_height
        else:
            self._line_height = float(line_height)

        # ALPHABET
        if alphabet is None:
            if self._default_alphabet is None:
                self._alphabet = None
            else:
                self._alphabet = self._default_alphabet
        else:
            self._alphabet = str(alphabet)
            self._alpha_solo = _re.compile(f"[{self._alphabet}]")
            self._alpha_plus = _re.compile(f"[{self._alphabet}]+")

        # LOAD FONT
        self._font = _font.Font(self._font_face, self._font_size)
        self._x_height = self._font.calculate_height("x")
        self._half_space_width = self._font.calculate_width(" ") / 2

        # INITIALIZE CHARACTERS AND ZONES
        self._characters, self._zones = self._initialize_text_block()

        # SET REMAINING TEXTBLOCK COORDINATES
        self._x_br = max([line[-1].x_br for line in self._characters])
        self._y_tl = self._characters[0][0].y_tl
        self._y_br = self._y_tl + self.n_rows * self._line_height

    def __repr__(self):
        if len(self._characters[0]) >= 20:
            short_line1 = "".join(map(str, self._characters[0][:16]))
            return f"TextBlock[{short_line1}...]"
        line1 = "".join(map(str, self._characters[0]))
        return f"TextBlock[{line1}]"

    def __getitem__(self, id):
        """

        Subsetting a TextBlock object with a key of the form x:y:z returns
        characters y to z on row x as an InterestArea.

        """
        if isinstance(id, str):
            if id in self._zones:
                return self._zones[id]
            rse = id.split(":")
            try:
                r, s, e = int(rse[0]), int(rse[1]), int(rse[2])
            except:
                raise KeyError("Invalid InterestArea ID")
        elif isinstance(id, slice):
            r, s, e = id.start, id.stop, id.step
        return InterestArea(self._characters[r][s:e], f"{r}:{s}:{e}")

    def __len__(self):
        return sum([len(line) for line in self._characters])

    def __iter__(self):
        """

        Iterating over a TextBlock object yields each character in the text.

        """
        for line in self._characters:
            for char in line:
                yield char

    @property
    def text(self) -> str:
        """String representation of the text"""
        return " ".join(["".join(map(str, line)) for line in self._characters])

    @property
    def position(self) -> tuple:
        """Position of the `TextBlock`"""
        return self._x_tl, self._first_baseline

    @property
    def font_face(self) -> str:
        """Name of the font"""
        return self._font_face

    @property
    def font_size(self) -> float:
        """Font size in points"""
        return self._font_size

    @property
    def line_height(self) -> float:
        """Line height in points"""
        return self._line_height

    @property
    def alphabet(self) -> str:
        """Characters that are considered alphabetical"""
        return self._alphabet

    @property
    def n_rows(self) -> int:
        """Number of rows in the text (i.e. the number of lines)"""
        return len(self._characters)

    @property
    def n_cols(self) -> int:
        """Number of columns in the text (i.e. the number of characters in the widest line)"""
        return max([len(row) for row in self._characters])

    @property
    def line_positions(self) -> list:
        """Y-coordinates of the center of each line of text"""
        return [line[0].y for line in self._characters]

    @property
    def word_centers(self) -> list:
        """XY-coordinates of the center of each word"""
        return [word.center for word in self.words()]

    ################
    # PUBLIC METHODS
    ################

    def zones(self):
        """

        Iterate over each marked up zone as an `InterestArea`.

        """
        for _, zone in self._zones.items():
            yield zone

    def which_zone(self, fixation):
        """

        Return the marked-up zone that the fixation falls inside as an
        `InterestArea`.

        """
        for zone in self.zones():
            if fixation in zone:
                return zone
        return None

    def lines(self):
        """

        Iterate over each line as an `InterestArea`.

        """
        for r, line in enumerate(self._characters):
            yield InterestArea(line, f"{r}:{0}:{len(line)}")

    def which_line(self, fixation):
        """

        Return the line that the fixation falls inside as an `InterestArea`.

        """
        for line in self.lines():
            if fixation in line:
                return line
        return None

    def words(self, pattern=None, line_n=None, add_padding=True):
        """

        Iterate over each word as an `InterestArea`. Optionally, you can
        supply a regex pattern to define what constitutes a word or to pick
        out specific words. For example, `r'\\b[Tt]he\\b'` gives you all
        occurrences of the word *the* or `'[a-z]+ing'` gives you all words
        ending with *-ing*. `add_padding` adds half of the width of a space
        character to the left and right edges of the word's bounding box, so
        that fixations that fall on a space between two words will at least
        fall into one of the two words' bounding boxes.

        """
        if pattern is None:
            pattern = self._alpha_plus
        else:
            pattern = _re.compile(pattern)
        if add_padding:
            padding = self._half_space_width
        else:
            padding = 0
        word_i = 0
        for r, line in enumerate(self._characters):
            if line_n is not None and r != line_n:
                continue
            line_str = "".join(map(str, line))
            for word in pattern.findall(line_str):
                s = line_str.find(word)
                e = s + len(word)
                line_str = line_str.replace(word, "#" * len(word), 1)
                yield InterestArea(
                    self._characters[r][s:e], f"{r}:{s}:{e}", padding=padding
                )
                word_i += 1

    def which_word(self, fixation, pattern=None, line_n=None, add_padding=True):
        """

        Return the word that the fixation falls inside as an `InterestArea`.
        For the meaning of `pattern` and `add_padding` see
        `TextBlock.words()`.

        """
        for word in self.words(pattern, line_n, add_padding):
            if fixation in word:
                return word
        return None

    def characters(self, line_n=None, alphabetical_only=True):
        """

        Iterate over each character as an `InterestArea`.

        """
        char_i = 0
        for r, line in enumerate(self._characters):
            if line_n is not None and r != line_n:
                continue
            for s, char in enumerate(line):
                if alphabetical_only and not self._alpha_solo.match(str(char)):
                    continue
                yield InterestArea([char], f"{r}:{s}:{s+1}")
                char_i += 1

    def which_character(self, fixation, line_n=None, alphabetical_only=True):
        """

        Return the character that the fixation falls inside as an
        `InterestArea`.

        """
        for character in self.characters(line_n, alphabetical_only):
            if fixation in character:
                return character
        return None

    def ngrams(self, n, line_n=None, alphabetical_only=True):
        """

        Iterate over each ngram, for given n, as an `InterestArea`.

        """
        ngram_i = 0
        for r, line in enumerate(self._characters):
            if line_n is not None and r != line_n:
                continue
            for s in range(len(line) - (n - 1)):
                e = s + n
                if alphabetical_only and not self._alpha_plus.fullmatch(
                    "".join(map(str, line[s:e]))
                ):
                    continue
                ngram = InterestArea(line[s:e], f"{r}:{s}:{e}")
                yield ngram
                ngram_i += 1

    # No which_ngram() method because, by definition, a fixation is inside
    # multiple ngrams.

    #################
    # PRIVATE METHODS
    #################

    def _serialize(self):
        """

        Returns the `TextBlock`'s initialization arguments as a dictionary for
        serialization.

        """
        data = {
            "position": self.position,
            "font_face": self.font_face,
            "font_size": self.font_size,
            "line_height": self.line_height,
        }
        if self.alphabet is not None:
            data["alphabet"] = self.alphabet
        data["text"] = self._text
        return data

    def _initialize_text_block(self):
        """

        Parses out any marked up interest areas from the text and converts and
        stores every character as a Character object with the appropriate
        character width.

        """
        characters, zones = [], {}
        baseline = self._first_baseline
        # midline is half an x-height higher than the baseline
        midline = baseline - self._x_height / 2
        # top of bounding box is half a line height above the midline
        y_tl = midline - self._line_height / 2
        for r, line in enumerate(self._text):

            # PARSE AND STRIP OUT INTEREST AREA ZONES FROM THIS LINE
            for zone_markup, zone_text, zone_id in self._zone_markup.findall(line):
                if zone_id in zones:
                    raise ValueError(
                        f'The zone ID "{zone_id}" has been used more than once.'
                    )
                c = line.find(zone_markup)
                # record row index, column index, and length of zone
                zones[zone_id] = (r, c, len(zone_text))
                # replace the marked up zone with the unmarked up text
                line = line.replace(zone_markup, zone_text)

            # CREATE THE SET OF CHARACTER OBJECTS FOR THIS LINE
            characters_line = []
            # x_tl of first character bounding box on this line
            x_tl = self.x_tl
            # y_br of all character bounding boxes on this line
            y_br = y_tl + self._line_height
            for char in line:
                char_width = self._font.calculate_width(char)
                x_br = x_tl + char_width
                characters_line.append(
                    Character(char, x_tl, y_tl, x_br, y_br, baseline)
                )
                x_tl = x_br  # next x_tl is x_br
            characters.append(characters_line)
            y_tl += self._line_height
            baseline += self._line_height

        # SET UP AND STORE THE ZONED INTEREST AREAS BASED ON THE INDICES
        # STORED EARLIER. This needs to be done in two steps because IAs can't
        # be created until character positions are known.
        for zone_id, (r, c, length) in zones.items():
            zones[zone_id] = InterestArea(characters[r][c : c + length], zone_id)

        return characters, zones



"""

Defines the `Image`, `Figure`, and `Booklet` objects, which are used to create
visualizations.

"""


from os import path as _path
import cairocffi as _cairo
from . import _color, _font


_FONT_FACE = "Arial"
_FONT_SIZE = 8


def set_default_font(font_face=None, font_size=None):
    """

    Set the default font face and/or size that will be used in figure captions
    and image annotations. This selection can be overridden on a per-image or
    per-figure basis. If no font is set, Eyekit defaults to 8pt Arial.

    """
    global _FONT_FACE, _FONT_SIZE
    if font_face is not None:
        _FONT_FACE = str(font_face)
    if font_size is not None:
        _FONT_SIZE = float(font_size)


class Image(object):

    """

    The Image class is used to create visualizations of text blocks and
    fixation sequences, and it provides methods for drawing various kinds of
    annotation. The general usage pattern is:

    ```python
    img = eyekit.vis.Image(1920, 1080)
    img.draw_text_block(txt)
    img.draw_fixation_sequence(seq)
    img.save('image.pdf')
    ```

    """

    def __init__(self, screen_width: int, screen_height: int):
        """Initialized with:

        - `screen_width` Width of the screen in pixels.
        - `screen_height` Height of the screen in pixels.
        """
        self.screen_width = int(screen_width)
        self.screen_height = int(screen_height)
        self._caption = None
        self._caption_font_face = _FONT_FACE
        self._caption_font_size = _FONT_SIZE
        self._background_color = (1, 1, 1)
        self._components = []
        self._text_extents = None

    ################
    # PUBLIC METHODS
    ################

    def set_caption(self, caption, font_face=None, font_size=None):
        """

        Set the image's caption, which will be shown above the image if you
        place it inside a `Figure`.

        """
        self._caption = str(caption)
        if font_face is not None:
            self._caption_font_face = str(font_face)
        if font_size is not None:
            self._caption_font_size = float(font_size)

    def set_background_color(self, color):
        """

        Set the background color of the image. By default the background color
        is white. If `color` is set to `None`, the background will be
        transparent.

        """
        if color is None:
            self._background_color = None
        else:
            self._background_color = _color_to_rgb(color)

    def draw_text_block(self, text_block, color="black"):
        """

        Draw a `eyekit.text.TextBlock` on the image. `color` sets the color of
        the text.

        """
        if self._text_extents is None:
            self._text_extents = [
                text_block.x_tl,
                text_block.y_tl,
                text_block.x_br,
                text_block.y_br,
            ]
        else:
            if text_block.x_tl < self._text_extents[0]:
                self._text_extents[0] = text_block.x_tl
            if text_block.y_tl < self._text_extents[1]:
                self._text_extents[1] = text_block.y_tl
            if text_block.x_br > self._text_extents[2]:
                self._text_extents[2] = text_block.x_br
            if text_block.y_br > self._text_extents[3]:
                self._text_extents[3] = text_block.y_br
        rgb_color = _color_to_rgb(color)
        for line in text_block.lines():
            arguments = {
                "x": line.x_tl,
                "y": line.baseline,
                "text": line.text,
                "font": text_block._font,
                "color": rgb_color,
            }
            self._add_component(_draw_text, arguments)

    def draw_text_block_heatmap(self, text_block, distribution, color="red"):
        """

        Draw a `eyekit.text.TextBlock` on the image along with an associated
        distribution, which is represented in heatmap form. This is can be
        used to visualize the output from `eyekit.analysis.duration_mass()`.
        `color` determines the color of the heatmap.

        """
        rgb_color = _color_to_rgb(color)
        n = (text_block.n_cols - distribution.shape[1]) + 1
        distribution /= distribution.max()
        subcell_height = text_block.line_height / n
        levels = [subcell_height * i for i in range(n)]
        level = 0
        for ngram in text_block.ngrams(n):
            r, s, e = ngram.id.split(":")
            if level == n:
                level = 0
            p = distribution[(int(r), int(s))]
            cell_color = _pseudo_alpha(rgb_color, opacity=p)
            arguments = {
                "x": ngram.x_tl,
                "y": ngram.y_tl + subcell_height * level,
                "width": ngram.width,
                "height": subcell_height,
                "stroke_width": None,
                "color": None,
                "fill_color": cell_color,
                "dashed": False,
            }
            self._add_component(_draw_rectangle, arguments)
            level += 1
        start_x = text_block.x_tl
        end_x = text_block.x_br
        for line_n, line in enumerate(text_block.lines()):
            if line_n == 0:
                continue
            y = line.y_tl
        self.draw_text_block(text_block)

    def draw_fixation_sequence(
        self,
        fixation_sequence,
        show_saccades=True,
        show_discards=False,
        color="black",
        discard_color="gray",
        number_fixations=False,
    ):
        """

        Draw a `eyekit.fixation.FixationSequence` on the image. Optionally, you can
        choose whether or not to display saccade lines and discarded fixations, and
        which colors to use. `number_fixations` is not yet implemented.

        """
        rgb_color = _color_to_rgb(color)
        rgb_discard_color = _color_to_rgb(discard_color)
        if show_discards:
            if show_saccades:
                path = [
                    fixation.xy for fixation in fixation_sequence.iter_with_discards()
                ]
                arguments = {
                    "path": path,
                    "stroke_width": 0.5,
                    "color": rgb_color,
                    "dashed": False,
                }
                self._add_component(_draw_line, arguments)
            for fixation in fixation_sequence.iter_with_discards():
                f_color = rgb_discard_color if fixation.discarded else rgb_color
                arguments = {
                    "x": fixation.x,
                    "y": fixation.y,
                    "radius": _duration_to_radius(fixation.duration),
                    "color": None,
                    "stroke_width": None,
                    "dashed": False,
                    "fill_color": f_color,
                }
                self._add_component(_draw_circle, arguments)
        else:
            if show_saccades:
                path = [
                    fixation.xy
                    for fixation in fixation_sequence.iter_without_discards()
                ]
                arguments = {
                    "path": path,
                    "stroke_width": 0.5,
                    "color": rgb_color,
                    "dashed": False,
                }
                self._add_component(_draw_line, arguments)
            for fixation in fixation_sequence.iter_without_discards():
                arguments = {
                    "x": fixation.x,
                    "y": fixation.y,
                    "radius": _duration_to_radius(fixation.duration),
                    "color": None,
                    "stroke_width": None,
                    "dashed": False,
                    "fill_color": rgb_color,
                }
                self._add_component(_draw_circle, arguments)

    def draw_sequence_comparison(
        self,
        reference_sequence,
        fixation_sequence,
        color_match="black",
        color_mismatch="red",
    ):
        """

        Draw a `eyekit.fixation.FixationSequence` on the image with the
        fixations colored according to whether or not they match a reference
        sequence in terms of the y-coordinate. This is mostly useful for
        comparing the outputs of two different drift correction algorithms.

        """
        rgb_color_match = _color_to_rgb(color_match)
        rgb_color_mismatch = _color_to_rgb(color_mismatch)
        path = [fixation.xy for fixation in fixation_sequence.iter_with_discards()]
        arguments = {
            "path": path,
            "stroke_width": 0.5,
            "color": rgb_color_match,
            "dashed": False,
        }
        self._add_component(_draw_line, arguments)
        for reference_fixation, fixation in zip(
            reference_sequence.iter_with_discards(),
            fixation_sequence.iter_with_discards(),
        ):
            if reference_fixation.y == fixation.y:
                color = rgb_color_match
            else:
                color = rgb_color_mismatch
            arguments = {
                "x": fixation.x,
                "y": fixation.y,
                "radius": _duration_to_radius(fixation.duration),
                "color": None,
                "stroke_width": None,
                "dashed": False,
                "fill_color": color,
            }
            self._add_component(_draw_circle, arguments)

    def draw_line(self, start_xy, end_xy, color="black", stroke_width=1, dashed=False):
        """

        Draw an arbitrary line on the image from `start_xy` to `end_xy`.
        `stroke_width` is set in points for vector output or pixels for PNG
        output.

        """
        rgb_color = _color_to_rgb(color)
        arguments = {
            "path": [start_xy, end_xy],
            "stroke_width": stroke_width,
            "color": rgb_color,
            "dashed": dashed,
        }
        self._add_component(_draw_line, arguments)

    def draw_circle(
        self, x, y, radius, color="black", stroke_width=1, dashed=False, fill_color=None
    ):
        """

        Draw an arbitrary circle on the image centered at `x`, `y` and with
        some `radius`. `stroke_width` is set in points for vector output or
        pixels for PNG output.

        """
        rgb_color = _color_to_rgb(color) if color else None
        rgb_fill_color = _color_to_rgb(fill_color) if fill_color else None
        arguments = {
            "x": x,
            "y": y,
            "radius": radius,
            "color": rgb_color,
            "stroke_width": stroke_width,
            "dashed": dashed,
            "fill_color": rgb_fill_color,
        }
        self._add_component(_draw_circle, arguments)

    def draw_rectangle(
        self,
        rect,
        y=None,
        width=None,
        height=None,
        color="black",
        stroke_width=1,
        dashed=False,
        fill_color=None,
    ):
        """

        Draw an arbitrary rectangle on the image. `rect` should be a tuple
        specifying x, y, width, and height, or these four values can be passed
        in separately as the first four arguments. `stroke_width` is set in
        points for vector output or pixels for PNG output.

        """
        rgb_color = _color_to_rgb(color) if color else None
        rgb_fill_color = _color_to_rgb(fill_color) if fill_color else None
        if isinstance(rect, tuple) and len(rect) == 4:
            x, y, width, height = rect
        else:
            x = rect
        arguments = {
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "color": rgb_color,
            "stroke_width": stroke_width,
            "dashed": dashed,
            "fill_color": rgb_fill_color,
        }
        self._add_component(_draw_rectangle, arguments)

    def draw_annotation(
        self, x, y, text, font_face=None, font_size=None, color="black"
    ):
        """

        Draw arbitrary text on the image located at `x`, `y`. `font_size` is
        set in points for vector output or pixels for PNG output.

        """
        if font_face is None:
            font_face = _FONT_FACE
        if font_size is None:
            font_size = _FONT_SIZE
        font = _font.Font(font_face, font_size)
        rgb_color = _color_to_rgb(color)
        arguments = {
            "x": x,
            "y": y,
            "text": text,
            "font": font,
            "color": rgb_color,
            "annotation": True,
        }
        self._add_component(_draw_text, arguments)

    def save(self, output_path, width=150, crop_margin=None):
        """

        Save the image to some `output_path`. Images can be saved as .pdf,
        .eps, .svg, or .png. `width` only applies to the vector formats and
        determines the millimeter width of the output file; PNG images are
        saved at actual pixel size. If you set a crop margin, the image will
        be cropped to the size of the `eyekit.text.TextBlock` plus the
        specified margin. Margins are specified in millimeters (PDF, EPS, SVG)
        or pixels (PNG).

        """
        image_format = _path.splitext(output_path)[1][1:].upper()
        if image_format not in ["PDF", "EPS", "SVG", "PNG"]:
            raise ValueError(
                "Unrecognized format. Use .pdf, .eps, or .svg for vector output, or .png for raster output."
            )
        image_width = _mm_to_pts(width)
        surface, context, scale = self._make_surface(
            output_path, image_format, image_width, crop_margin
        )
        self._render_background(context)
        self._render_components(context, scale)
        if image_format == "PNG":
            surface.write_to_png(output_path)
        surface.finish()

    #################
    # PRIVATE METHODS
    #################

    def _add_component(self, func, arguments):
        """

        Add a component to the stack. This should be a draw_ function and its
        argumments. This function will be called with its arguments at save
        time.

        """
        self._components.append((func, arguments))

    def _make_surface(self, output_path, image_format, image_width, crop_margin):
        """

        Make the relevant Cairo surface and context with appropriate sizing.

        """
        text_width = self._text_extents[2] - self._text_extents[0]
        text_height = self._text_extents[3] - self._text_extents[1]
        if image_format == "PNG":
            scale = 1
            if crop_margin is None:
                image_width = self.screen_width
                image_height = self.screen_height
            else:
                image_width = int(text_width + crop_margin * 2)
                image_height = int(text_height + crop_margin * 2)
            surface = _cairo.ImageSurface(
                _cairo.FORMAT_ARGB32, image_width, image_height
            )
        else:
            if crop_margin is None:
                scale = image_width / self.screen_width
                image_height = self.screen_height * scale
            else:
                crop_margin = _mm_to_pts(crop_margin)
                if crop_margin > image_width / 3:
                    raise ValueError(
                        "The crop margin set on this image is too large for the image width. Increase the image width or decrease the crop margin."
                    )
                scale = (image_width - crop_margin * 2) / text_width
                image_height = text_height * scale + crop_margin * 2
            if image_format == "PDF":
                surface = _cairo.PDFSurface(output_path, image_width, image_height)
            elif image_format == "EPS":
                surface = _cairo.PSSurface(output_path, image_width, image_height)
                surface.set_eps(True)
            elif image_format == "SVG":
                surface = _cairo.SVGSurface(output_path, image_width, image_height)
            surface.set_device_scale(scale, scale)
        context = _cairo.Context(surface)
        if crop_margin is not None:
            crop_margin = crop_margin / scale
            context.translate(
                -self._text_extents[0] + crop_margin,
                -self._text_extents[1] + crop_margin,
            )
        return surface, context, scale

    def _render_background(self, context):
        """

        Render the background color.

        """
        if self._background_color is not None:
            with context:
                context.set_source_rgb(*self._background_color)
                context.paint()

    def _render_components(self, context, scale):
        """

        Render all components in the components stack (functions and function
        arguments that must be called in sequence).

        """
        for func, arguments in self._components:
            with context:
                func(context, scale, **arguments)

    def _render_to_figure(self, context, scale):
        """

        Render the image to a figure panel.

        """
        self._render_background(context)
        self._render_components(context, scale)


class Figure(object):

    """

    The Figure class is used to combine one or more images into a
    publication-ready figure. The general usage pattern is:

    ```python
    fig = eyekit.vis.Figure(1, 2)
    fig.add_image(img1)
    fig.add_image(img2)
    fig.save('figure.pdf')
    ```

    """

    def __init__(self, n_rows: int = 1, n_cols: int = 1):
        """Initialized with:

        - `n_rows` Number of rows in the figure.
        - `n_cols` Number of columns in the figure.
        """
        self._n_rows = int(n_rows)
        if self._n_rows <= 0:
            raise ValueError("Invalid number of rows")
        self._n_cols = int(n_cols)
        if self._n_rows <= 0:
            raise ValueError("Invalid number of columns")
        self._grid = [[None] * self._n_cols for _ in range(self._n_rows)]
        self._crop_margin = None
        self._lettering = True
        self._letter_font_face = _FONT_FACE + " bold"
        self._letter_font_size = _FONT_SIZE
        self._v_padding = 4
        self._h_padding = 4
        self._e_padding = 1

    ################
    # PUBLIC METHODS
    ################

    def set_crop_margin(self, crop_margin):
        """

        Set the crop margin of the embedded images. Each image in the figure
        will be cropped to the size and positioning of the most extreme text
        block extents, plus the specified margin. This has the effect of
        zooming in to all images in a consistent way – maintaining the aspect
        ratio and relative positioning of the text blocks across images.
        Margins are specified in figure millimeters.

        """
        self._crop_margin = _mm_to_pts(crop_margin)

    def set_lettering(self, lettering=True, font_face=None, font_size=None):
        """

        By default, each image caption is prefixed with a letter, **(A)**,
        **(B)**, **(C)**, etc. If you want to turn this off, call
        ```Figure.set_lettering(False)``` prior to saving.

        """
        self._lettering = bool(lettering)
        if font_face is not None:
            self._letter_font_face = str(font_face)
        if font_size is not None:
            self._letter_font_size = float(font_size)

    def set_padding(self, vertical=None, horizontal=None, edge=None):
        """

        Set the vertical or horizontal padding between images or the padding
        around the edge of the figure. Padding is expressed in millimeters. By
        default, the vertical and horizontal padding between images is 4mm and
        the edge padding is 1mm.

        """
        if vertical is not None:
            self._v_padding = float(vertical)
        if horizontal is not None:
            self._h_padding = float(horizontal)
        if edge is not None:
            self._e_padding = float(edge)

    def add_image(self, image, row=None, col=None):
        """

        Add an `Image` to the figure. If a row and column index is specified,
        the image is placed in that position. Otherwise, `image` is placed in
        the next available position.

        """
        if not isinstance(image, Image):
            raise TypeError("image should be of type Image.")
        if row is None or col is None:
            row, col = self._next_available_cell(row, col)
        if row >= self._n_rows or col >= self._n_cols:
            raise ValueError("Row or column index is not inside the grid.")
        self._grid[row][col] = image

    def save(self, output_path, width=150):
        """

        Save the figure to some `output_path`. Figures can be saved as .pdf,
        .eps, or .svg. `width` determines the millimeter width of the output
        file.

        """
        figure_format = _path.splitext(output_path)[1][1:].upper()
        if figure_format not in ["PDF", "EPS", "SVG"]:
            raise ValueError("Unrecognized format. Use .pdf, .eps, or .svg.")
        figure_width = _mm_to_pts(width)
        layout, components, height, text_block_extents = self._make_layout(figure_width)
        surface, context = self._make_surface(
            output_path, figure_format, figure_width, height
        )
        self._render_background(context)
        self._render_images(surface, layout, text_block_extents)
        self._render_components(context, components)
        surface.finish()

    #################
    # PRIVATE METHODS
    #################

    def _next_available_cell(self, row_i=None, col_i=None):
        """

        Get the indices of the next available cell in the figure, optionally
        in a specific row or column.

        """
        for i, row in enumerate(self._grid):
            if row_i is not None and row_i != i:
                continue
            for j, cell in enumerate(row):
                if col_i is not None and col_i != j:
                    continue
                if cell is None:
                    return i, j
        raise ValueError(
            "Cannot add image to the figure because there are no available positions. Make a new Figure with more rows or columns, or specify a specific row and column index to overwrite the image that is currently in that position."
        )

    def _make_surface(self, output_path, figure_format, figure_width, figure_height):
        """

        Make the relevant Cairo surface and context with appropriate sizing.

        """
        if figure_format == "PDF":
            surface = _cairo.PDFSurface(output_path, figure_width, figure_height)
        elif figure_format == "EPS":
            surface = _cairo.PSSurface(output_path, figure_width, figure_height)
            surface.set_eps(True)
        elif figure_format == "SVG":
            surface = _cairo.SVGSurface(output_path, figure_width, figure_height)
        context = _cairo.Context(surface)
        return surface, context

    def _max_text_block_extents(self):
        """

        Calculate the maximum text block extents from all images in the
        figure.

        """
        x_tl, y_tl, x_br, y_br = 999999, 999999, 0, 0
        fallback = None
        for row in self._grid:
            for image in row:
                if isinstance(image, Image) and image._text_extents:
                    if image._text_extents[0] < x_tl:
                        x_tl = image._text_extents[0]
                    if image._text_extents[1] < y_tl:
                        y_tl = image._text_extents[1]
                    if image._text_extents[2] > x_br:
                        x_br = image._text_extents[2]
                    if image._text_extents[3] > y_br:
                        y_br = image._text_extents[3]
                elif isinstance(image, Image):
                    fallback = [0, 0, image.screen_width, image.screen_height]
        if x_tl < x_br:
            return x_tl, y_tl, x_br - x_tl, y_br - y_tl
        if fallback is None:
            raise ValueError(
                "Cannot make figure because no images have been added. Use Figure.add_image()"
            )
        return tuple(fallback)

    def _make_layout(self, figure_width):
        """

        Figure out the layout of the figure, and append all the relevant
        components to the stack for rendering. This needs to be done in two
        steps in order to determine the appropriate figure height.

        """
        layout, components = [], []
        letter_index = 65  # 65 == A, etc...
        text_block_extents = self._max_text_block_extents()
        v_padding = _mm_to_pts(self._v_padding)
        h_padding = _mm_to_pts(self._h_padding)
        e_padding = _mm_to_pts(self._e_padding)
        letter_font = _font.Font(self._letter_font_face, self._letter_font_size)
        y = e_padding
        for row in self._grid:
            x = e_padding
            tallest_in_row = 0
            caption_count = sum(
                [bool(image._caption) for image in row if isinstance(image, Image)]
            )
            if self._lettering or caption_count > 0:
                # row contains captions, so make some space
                y += self._letter_font_size * 2
            n_cols = len(row)
            cell_width = (
                figure_width - 2 * e_padding - (n_cols - 1) * h_padding
            ) / n_cols
            for image in row:
                if image is None:
                    x += cell_width + h_padding
                    continue
                if self._crop_margin is None:
                    scale = cell_width / image.screen_width
                    aspect_ratio = image.screen_width / image.screen_height
                else:
                    scale = (cell_width - self._crop_margin * 2) / text_block_extents[2]
                    aspect_ratio = text_block_extents[2] / text_block_extents[3]
                cell_height = cell_width / aspect_ratio
                if cell_height > tallest_in_row:
                    tallest_in_row = cell_height
                caption_advance = 0
                if self._lettering:
                    letter_prefix = f"({chr(letter_index)}) "
                    arguments = {
                        "x": x,
                        "y": y - self._letter_font_size,
                        "text": letter_prefix,
                        "font": letter_font,
                        "color": (0, 0, 0),
                    }
                    components.append((_draw_text, arguments))
                    caption_advance += letter_font.calculate_width(letter_prefix)
                if image._caption:
                    caption_font = _font.Font(
                        image._caption_font_face, image._caption_font_size
                    )
                    arguments = {
                        "x": x + caption_advance,
                        "y": y - self._letter_font_size,
                        "text": image._caption,
                        "font": caption_font,
                        "color": (0, 0, 0),
                    }
                    components.append((_draw_text, arguments))
                layout.append((image, x, y, cell_width, cell_height, scale))
                arguments = {
                    "x": x,
                    "y": y,
                    "width": cell_width,
                    "height": cell_height,
                    "color": (0, 0, 0),
                    "stroke_width": 1,
                    "dashed": False,
                    "fill_color": None,
                }
                components.append((_draw_rectangle, arguments))
                x += cell_width + h_padding
                letter_index += 1
            y += tallest_in_row + v_padding
        figure_height = y - (v_padding - e_padding)
        return layout, components, figure_height, text_block_extents

    def _render_background(self, context):
        """

        Render the background color.

        """
        with context:
            context.set_source_rgb(1, 1, 1)
            context.paint()

    def _render_images(self, surface, layout, text_block_extents):
        """

        Render all images. This creates a separate subsurface for each image
        to be rendered to.

        """
        min_x, min_y, max_width, max_height = text_block_extents
        aspect_ratio = max_width / max_height
        for image, x, y, width, height, scale in layout:
            subsurface = surface.create_for_rectangle(x, y, width, height)
            subsurface.set_device_scale(scale, scale)
            context = _cairo.Context(subsurface)
            if self._crop_margin is not None:
                context.translate(
                    -min_x + self._crop_margin / scale,
                    -min_y + self._crop_margin / (scale * aspect_ratio),
                )
            image._render_to_figure(context, scale)

    def _render_components(self, context, components):
        """

        Render all components in the components stack (functions and function
        arguments that must be called in sequence).

        """
        for func, arguments in components:
            with context:
                func(context, 1, **arguments)

    def _render_to_booklet(self, surface, context, width):
        """

        Render the figure to a booklet page.

        """
        layout, components, height, text_block_extents = self._make_layout(width)
        self._render_background(context)
        self._render_images(surface, layout, text_block_extents)
        self._render_components(context, components)


class Booklet(object):

    """

    The Booklet class is used to combine one or more figures into a multipage
    PDF booklet. The general usage pattern is:

    ```python
    booklet = eyekit.vis.Booklet()
    booklet.add_figure(fig1)
    booklet.add_figure(fig2)
    booklet.save('booklet.pdf')
    ```

    """

    def __init__(self):
        self._figures = []

    def add_figure(self, figure):
        """

        Add a `Figure` to a new page in the booklet.

        """
        self._figures.append(figure)

    def save(self, output_path, width=210, height=297):
        """

        Save the booklet to some `output_path`. Booklets can only be saved as
        .pdf. `width` and `height` determine the millimeter sizing of the
        booklet pages, which defaults to A4 (210x297mm).

        """
        if _path.splitext(output_path)[1][1:].lower() != "pdf":
            raise ValueError("Books must be saved in PDF format.")
        page_width = _mm_to_pts(width)
        page_height = _mm_to_pts(height)
        surface = _cairo.PDFSurface(output_path, page_width, page_height)
        context = _cairo.Context(surface)
        for figure in self._figures:
            figure._render_to_booklet(surface, context, page_width)
            surface.show_page()
        surface.finish()


################
# DRAW FUNCTIONS
################


def _draw_line(context, scale, path, color, stroke_width, dashed):
    context.set_source_rgb(*color)
    context.set_line_width(stroke_width / scale)
    context.move_to(*path[0])
    if dashed:
        context.set_dash([10, 4])
    for end_xy in path[1:]:
        context.line_to(*end_xy)
    context.stroke()


def _draw_circle(context, scale, x, y, radius, color, stroke_width, dashed, fill_color):
    context.arc(x, y, radius, 0, 6.283185307179586)
    if color and stroke_width:
        context.set_source_rgb(*color)
        context.set_line_width(stroke_width / scale)
        if dashed:
            context.set_dash([12, 4])
        if fill_color:
            context.stroke_preserve()
        else:
            context.stroke()
    if fill_color:
        context.set_source_rgb(*fill_color)
        context.fill()


def _draw_rectangle(
    context, scale, x, y, width, height, color, stroke_width, dashed, fill_color
):
    context.rectangle(x, y, width, height)
    if color and stroke_width:
        context.set_source_rgb(*color)
        context.set_line_width(stroke_width / scale)
        if dashed:
            context.set_dash([12, 4])
        if fill_color:
            context.stroke_preserve()
        else:
            context.stroke()
    if fill_color:
        context.set_source_rgb(*fill_color)
        context.fill()


def _draw_text(context, scale, x, y, text, font, color, annotation=False):
    context.set_source_rgb(*color)
    context.set_font_face(font.toy_font_face)
    if annotation:
        context.set_font_size(font.size / scale)
    else:
        context.set_font_size(font.size)
    context.move_to(x, y)
    context.show_text(text)


##################
# HELPER FUNCTIONS
##################


def _duration_to_radius(duration):
    """

    Converts a millisecond duration to a pixel radius for plotting fixation
    circles so that the area of the circle corresponds to duration.

    """
    return (duration / 3.141592653589793) ** 0.5


def _mm_to_pts(mm):
    """

    Convert millimeters to points.

    """
    return mm / (25.4 / 72)


def _color_to_rgb(color):
    """

    Convert a color to RGB values in [0, 1]. Can take an RGB tuple in [0,
    255], a hex triplet, or a named color.

    """
    if isinstance(color, tuple) and len(color) == 3:
        return color[0] / 255, color[1] / 255, color[2] / 255
    if isinstance(color, str) and color[0] == "#":
        r, g, b = tuple(bytes.fromhex(color[1:]))
        return r / 255, g / 255, b / 255
    if color.lower() in _color.colors:
        color = _color.colors[color.lower()]
        return color[0] / 255, color[1] / 255, color[2] / 255
    return 0, 0, 0


def _pseudo_alpha(rgb, opacity):
    """

    Given an RGB value in [0, 1], return a new RGB value which blends in a
    certain amount of white to create a fake alpha effect. This allosws us to
    produce an alpha-like effect in EPS, which doesn't support transparency.

    """
    return tuple([value * opacity - opacity + 1 for value in rgb])



"""

Functions for calculating common analysis measures, such as total fixation
duration or initial landing position.

"""


import numpy as _np
from ._core import distance as _distance
from .fixation import FixationSequence as _FixationSequence, Fixation as _Fixation
from .text import TextBlock as _TextBlock, InterestArea as _InterestArea


def initial_fixation_duration(interest_areas, fixation_sequence):
    """

    Given an interest area or collection of interest areas, return the
    duration of the initial fixation on each interest area.

    """
    if isinstance(interest_areas, _InterestArea):
        interest_areas = [interest_areas]
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type FixationSequence")
    durations = {}
    for interest_area in interest_areas:
        if not isinstance(interest_area, _InterestArea):
            raise TypeError(f"{str(interest_area)} is not of type InterestArea")
        durations[interest_area.id] = 0
        for fixation in fixation_sequence.iter_without_discards():
            if fixation in interest_area:
                durations[interest_area.id] = fixation.duration
                break
    return durations


def total_fixation_duration(interest_areas, fixation_sequence):
    """

    Given an interest area or collection of interest areas, return the total
    fixation duration on each interest area.

    """
    if isinstance(interest_areas, _InterestArea):
        interest_areas = [interest_areas]
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type FixationSequence")
    durations = {}
    for interest_area in interest_areas:
        if not isinstance(interest_area, _InterestArea):
            raise TypeError(f"{str(interest_area)} is not of type InterestArea")
        durations[interest_area.id] = 0
        for fixation in fixation_sequence.iter_without_discards():
            if fixation in interest_area:
                durations[interest_area.id] += fixation.duration
    return durations


def gaze_duration(interest_areas, fixation_sequence):
    """

    Given an interest area or collection of interest areas, return the gaze
    duration on each interest area. Gaze duration is the sum duration of all
    fixations inside an interest area until the area is exited for the first
    time.

    """
    if isinstance(interest_areas, _InterestArea):
        interest_areas = [interest_areas]
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type FixationSequence")
    durations = {}
    for interest_area in interest_areas:
        if not isinstance(interest_area, _InterestArea):
            raise TypeError(f"{str(interest_area)} is not of type InterestArea")
        durations[interest_area.id] = 0
        for fixation in fixation_sequence.iter_without_discards():
            if fixation in interest_area:
                durations[interest_area.id] += fixation.duration
            elif durations[interest_area.id] > 0:
                break  # at least one previous fixation was inside the IA and this fixation is not, so break
    return durations


def initial_landing_position(interest_areas, fixation_sequence):
    """

    Given an interest area or collection of interest areas, return the initial
    landing position (expressed in character positions) on each interest area.
    Counting is from 1, so a 1 indicates the initial fixation landed on the
    first character and so forth.

    """
    if isinstance(interest_areas, _InterestArea):
        interest_areas = [interest_areas]
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type FixationSequence")
    positions = {}
    for interest_area in interest_areas:
        if not isinstance(interest_area, _InterestArea):
            raise TypeError(f"{str(interest_area)} is not of type InterestArea")
        positions[interest_area.id] = None
        for fixation in fixation_sequence.iter_without_discards():
            if fixation in interest_area:
                for position, char in enumerate(interest_area, 1):
                    if fixation in char:
                        positions[interest_area.id] = position
                        break
                break
    return positions


def initial_landing_x(interest_areas, fixation_sequence):
    """

    Given an interest area or collection of interest areas, return the initial
    landing position (expressed in pixel distance from the start of the
    interest area) on each interest area.

    """
    if isinstance(interest_areas, _InterestArea):
        interest_areas = [interest_areas]
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type FixationSequence")
    positions = {}
    for interest_area in interest_areas:
        if not isinstance(interest_area, _InterestArea):
            raise TypeError(f"{str(interest_area)} is not of type InterestArea")
        positions[interest_area.id] = None
        for fixation in fixation_sequence.iter_without_discards():
            if fixation in interest_area:
                for char in interest_area:
                    if fixation in char:
                        positions[interest_area.id] = fixation.x - interest_area.x_tl
                        break
                break
    return positions


def duration_mass(text_block, fixation_sequence, n=1, gamma=30):
    """

    Given a `eyekit.text.TextBlock` and `eyekit.fixation.FixationSequence`,
    distribute the durations of the fixations probabilistically across the
    `eyekit.text.TextBlock`. Specifically, the duration of fixation *f* is
    distributed over all characters *C* in its line according to the
    probability that the reader is "seeing" each character (see
    `p_characters_fixation()`), and this is summed over all fixations:

    $$\\sum_{f \\in F} p(C|f) \\cdot f_\\mathrm{dur}$$

    Returns a 2D Numpy array, the sum of which is equal to the total duration
    of all fixations. This can be passed to `eyekit.vis.Image.draw_text_block_heatmap()`
    for visualization. Duration mass reveals the parts of the text that
    received the most attention. Optionally, this can be performed over
    higher-level ngrams by setting `n` > 1.

    """
    if not isinstance(text_block, _TextBlock):
        raise TypeError("text_block should be of type TextBlock")
    if not isinstance(fixation_sequence, _FixationSequence):
        raise TypeError("fixation_sequence should be of type FixationSequence")
    shape = text_block.n_rows, text_block.n_cols - (n - 1)
    distribution = _np.zeros(shape, dtype=float)
    for fixation in fixation_sequence.iter_without_discards():
        distribution += (
            p_characters_fixation(text_block, fixation, n=n, gamma=gamma)
            * fixation.duration
        )
    return distribution


def p_characters_fixation(text_block, fixation, n=1, gamma=30):
    """

    Given a `eyekit.text.TextBlock` and `eyekit.fixation.Fixation`, calculate
    the probability that the reader is "seeing" each character in the text. We
    assume that the closer a character is to the fixation point, the greater
    the probability that the participant is "seeing" (i.e., processing) that
    character. Specifically, for a given fixation *f*, we compute a Gaussian
    distribution over all characters in the line according to:

    $$p(c|f) \\propto \\mathrm{exp} \\frac{ -\\mathrm{ED}(f_\\mathrm{pos}, c_\\mathrm{pos})^2 }{2\\gamma^2}$$

    where *γ* (`gamma`) is a free parameter controlling the rate at which
    probability decays with the Euclidean distance (ED) between the position
    of fixation *f* and the position of character *c*.

    Returns a 2D Numpy array representing a probability distribution over all
    characters, with all its mass confined to the line that the fixation
    occurred inside, and with greater mass closer to fixation points. This
    array can be passed to `eyekit.vis.Image.draw_text_block_heatmap()` for
    visualization. Optionally, this calculation can be performed over
    higher-level ngrams by setting `n` > 1.

    """
    if not isinstance(fixation, _Fixation):
        raise TypeError("fixation should be of type Fixation")
    line_positions = _np.array(text_block.line_positions, dtype=int)
    line_n = _np.argmin(abs(line_positions - fixation.y))
    shape = text_block.n_rows, text_block.n_cols - (n - 1)
    distribution = _np.zeros(shape, dtype=float)
    for ngram in text_block.ngrams(n, line_n=line_n):
        r, s, e = ngram.id.split(":")
        distance = _distance(fixation.xy, ngram.center)
        distribution[(int(r), int(s))] = _np.exp(-(distance ** 2) / (2 * gamma ** 2))
    return distribution / distribution.sum()



"""

These vertical drift correction algorithms were adapted from the Python code
provided at https://github.com/jwcarr/drift which is licensed under a creative
commons attribution license. It is expected that these implementations will
gradually diverge from the original work.

"""

import numpy as np

######################################################################
# CHAIN
#
# https://github.com/sascha2schroeder/popEye/
######################################################################


def chain(fixation_XY, text_block, x_thresh=192, y_thresh=32):
    """

    Chain consecutive fixations that are sufficiently close to each other, and
    then assign chains to their closest text lines. Default params:
    `x_thresh=192`, `y_thresh=32`.

    """
    line_Y = np.array(text_block.line_positions, dtype=int)
    dist_X = abs(np.diff(fixation_XY[:, 0]))
    dist_Y = abs(np.diff(fixation_XY[:, 1]))
    end_chain_indices = list(
        np.where(np.logical_or(dist_X > x_thresh, dist_Y > y_thresh))[0] + 1
    )
    end_chain_indices.append(len(fixation_XY))
    start_of_chain = 0
    for end_of_chain in end_chain_indices:
        mean_y = np.mean(fixation_XY[start_of_chain:end_of_chain, 1])
        line_i = np.argmin(abs(line_Y - mean_y))
        fixation_XY[start_of_chain:end_of_chain, 1] = line_Y[line_i]
        start_of_chain = end_of_chain
    return fixation_XY[:, 1]


######################################################################
# CLUSTER
#
# https://github.com/sascha2schroeder/popEye/
######################################################################


def cluster(fixation_XY, text_block):
    """

    Classify fixations into *m* clusters based on their Y-values, and then
    assign clusters to text lines in positional order. Requires SciPy.

    """
    try:
        from scipy.cluster.vq import kmeans2
    except ImportError:
        raise ImportError(
            'The cluster method requires SciPy. Run "pip install scipy" to use this method.'
        )
    line_Y = np.array(text_block.line_positions, dtype=int)
    fixation_Y = np.array(fixation_XY[:, 1], dtype=float)
    _, clusters = kmeans2(fixation_Y, line_Y, iter=100, minit="matrix", missing="warn")
    return line_Y[clusters]


######################################################################
# MERGE
#
# Špakov, O., Istance, H., Hyrskykari, A., Siirtola, H., & Räihä,
#   K.-J. (2019). Improving the performance of eye trackers with
#   limited spatial accuracy and low sampling rates for reading
#   analysis by heuristic fixation-to-word mapping. Behavior Research
#   Methods, 51(6), 2661–2687.
#
# https://doi.org/10.3758/s13428-018-1120-x
# https://github.com/uta-gasp/sgwm
######################################################################


def merge(fixation_XY, text_block, y_thresh=32, gradient_thresh=0.1, error_thresh=20):
    """

    Form a set of progressive sequences and then reduce the set to *m* by
    repeatedly merging those that appear to be on the same line. Merged
    sequences are then assigned to text lines in positional order. Default
    params: `y_thresh=32`, `gradient_thresh=0.1`, `error_thresh=20`.

    """
    line_Y = np.array(text_block.line_positions, dtype=int)
    diff_X = np.diff(fixation_XY[:, 0])
    dist_Y = abs(np.diff(fixation_XY[:, 1]))
    sequence_boundaries = list(
        np.where(np.logical_or(diff_X < 0, dist_Y > y_thresh))[0] + 1
    )
    sequence_starts = [0] + sequence_boundaries
    sequence_ends = sequence_boundaries + [len(fixation_XY)]
    sequences = [
        list(range(start, end)) for start, end in zip(sequence_starts, sequence_ends)
    ]
    for min_i, min_j, remove_constraints in [
        (3, 3, False),  # Phase 1
        (1, 3, False),  # Phase 2
        (1, 1, False),  # Phase 3
        (1, 1, True),  # Phase 4
    ]:
        while len(sequences) > len(line_Y):
            best_merger = None
            best_error = np.inf
            for i in range(len(sequences) - 1):
                if len(sequences[i]) < min_i:
                    continue  # first sequence too short, skip to next i
                for j in range(i + 1, len(sequences)):
                    if len(sequences[j]) < min_j:
                        continue  # second sequence too short, skip to next j
                    candidate_XY = fixation_XY[sequences[i] + sequences[j]]
                    gradient, intercept = np.polyfit(
                        candidate_XY[:, 0], candidate_XY[:, 1], 1
                    )
                    residuals = candidate_XY[:, 1] - (
                        gradient * candidate_XY[:, 0] + intercept
                    )
                    error = np.sqrt(sum(residuals ** 2) / len(candidate_XY))
                    if remove_constraints or (
                        abs(gradient) < gradient_thresh and error < error_thresh
                    ):
                        if error < best_error:
                            best_merger = (i, j)
                            best_error = error
            if best_merger is None:
                break  # no possible mergers, break while and move to next phase
            merge_i, merge_j = best_merger
            merged_sequence = sequences[merge_i] + sequences[merge_j]
            sequences.append(merged_sequence)
            del sequences[merge_j], sequences[merge_i]
    mean_Y = [fixation_XY[sequence, 1].mean() for sequence in sequences]
    ordered_sequence_indices = np.argsort(mean_Y)
    for line_i, sequence_i in enumerate(ordered_sequence_indices):
        fixation_XY[sequences[sequence_i], 1] = line_Y[line_i]
    return fixation_XY[:, 1]


######################################################################
# REGRESS
#
# Cohen, A. L. (2013). Software for the automatic correction of
#   recorded eye fixation locations in reading experiments. Behavior
#   Research Methods, 45(3), 679–683.
#
# https://doi.org/10.3758/s13428-012-0280-3
# https://blogs.umass.edu/rdcl/resources/
######################################################################


def regress(
    fixation_XY,
    text_block,
    slope_bounds=(-0.1, 0.1),
    offset_bounds=(-50, 50),
    std_bounds=(1, 20),
):
    """

    Find *m* regression lines that best fit the fixations and group fixations
    according to best fit regression lines, and then assign groups to text
    lines in positional order. Default params: `slope_bounds=(-0.1, 0.1)`,
    `offset_bounds=(-50, 50)`, `std_bounds=(1, 20)`. Requires SciPy.

    """
    try:
        from scipy.optimize import minimize
        from scipy.stats import norm
    except ImportError:
        raise ImportError(
            'The regress method requires SciPy. Run "pip install scipy" to use this method.'
        )
    line_Y = np.array(text_block.line_positions, dtype=int)
    density = np.zeros((len(fixation_XY), len(line_Y)))

    def fit_lines(params):
        k = slope_bounds[0] + (slope_bounds[1] - slope_bounds[0]) * norm.cdf(params[0])
        o = offset_bounds[0] + (offset_bounds[1] - offset_bounds[0]) * norm.cdf(
            params[1]
        )
        s = std_bounds[0] + (std_bounds[1] - std_bounds[0]) * norm.cdf(params[2])
        predicted_Y_from_slope = fixation_XY[:, 0] * k
        line_Y_plus_offset = line_Y + o
        for line_i in range(len(line_Y)):
            fit_Y = predicted_Y_from_slope + line_Y_plus_offset[line_i]
            density[:, line_i] = norm.logpdf(fixation_XY[:, 1], fit_Y, s)
        return -sum(density.max(axis=1))

    best_fit = minimize(fit_lines, [0, 0, 0], method="powell")
    fit_lines(best_fit.x)
    return line_Y[density.argmax(axis=1)]


######################################################################
# SEGMENT
#
# Abdulin, E. R., & Komogortsev, O. V. (2015). Person verification via
#   eye movement-driven text reading model, In 2015 IEEE 7th
#   International Conference on Biometrics Theory, Applications and
#   Systems. IEEE.
#
# https://doi.org/10.1109/BTAS.2015.7358786
######################################################################


def segment(fixation_XY, text_block):
    """

    Segment fixation sequence into *m* subsequences based on *m*–1 most-likely
    return sweeps, and then assign subsequences to text lines in chronological
    order.

    """
    line_Y = np.array(text_block.line_positions, dtype=int)
    diff_X = np.diff(fixation_XY[:, 0])
    saccades_ordered_by_length = np.argsort(diff_X)
    line_change_indices = saccades_ordered_by_length[: len(line_Y) - 1]
    current_line_i = 0
    for fixation_i in range(len(fixation_XY)):
        fixation_XY[fixation_i, 1] = line_Y[current_line_i]
        if fixation_i in line_change_indices:
            current_line_i += 1
    return fixation_XY[:, 1]


######################################################################
# SPLIT
######################################################################


def split(fixation_XY, text_block):
    """

    Split fixation sequence into subsequences based on best candidate return
    sweeps, and then assign subsequences to closest text lines. Requires
    SciPy.

    """
    try:
        from scipy.cluster.vq import kmeans2
    except ImportError:
        raise ImportError(
            'The split method requires SciPy. Run "pip install scipy" to use this method.'
        )
    line_Y = np.array(text_block.line_positions, dtype=int)
    diff_X = np.array(np.diff(fixation_XY[:, 0]), dtype=float).reshape(-1, 1)
    centers, clusters = kmeans2(diff_X, 2, iter=100, minit="++", missing="raise")
    sweep_marker = np.argmin(centers)
    end_line_indices = list(np.where(clusters == sweep_marker)[0] + 1)
    end_line_indices.append(len(fixation_XY))
    start_of_line = 0
    for end_of_line in end_line_indices:
        mean_y = np.mean(fixation_XY[start_of_line:end_of_line, 1])
        line_i = np.argmin(abs(line_Y - mean_y))
        fixation_XY[start_of_line:end_of_line] = line_Y[line_i]
        start_of_line = end_of_line
    return fixation_XY[:, 1]


######################################################################
# STRETCH
#
# Lohmeier, S. (2015). Experimental evaluation and modelling of the
#   comprehension of indirect anaphors in a programming language
#   (Master’s thesis). Technische Universität Berlin.
#
# http://www.monochromata.de/master_thesis/ma1.3.pdf
######################################################################


def stretch(
    fixation_XY, text_block, stretch_bounds=(0.9, 1.1), offset_bounds=(-50, 50)
):
    """

    Find a stretch factor and offset that results in a good alignment between
    the fixations and lines of text, and then assign the transformed fixations
    to the closest text lines. Default params: `stretch_bounds=(0.9, 1.1)`,
    `offset_bounds=(-50, 50)`. Requires SciPy.

    """
    try:
        from scipy.optimize import minimize
    except ImportError:
        raise ImportError(
            'The stretch method requires SciPy. Run "pip install scipy" to use this method.'
        )
    line_Y = np.array(text_block.line_positions, dtype=int)
    n = len(fixation_XY)
    fixation_Y = fixation_XY[:, 1]
    corrected_Y = np.zeros(n)

    def fit_lines(params):
        candidate_Y = fixation_Y * params[0] + params[1]
        for fixation_i in range(n):
            line_i = np.argmin(abs(line_Y - candidate_Y[fixation_i]))
            corrected_Y[fixation_i] = line_Y[line_i]
        return sum(abs(candidate_Y - corrected_Y))

    best_fit = minimize(
        fit_lines, [1, 0], method="powell", bounds=[stretch_bounds, offset_bounds]
    )
    fit_lines(best_fit.x)
    return corrected_Y


######################################################################
# WARP
#
# Python implementation of the Dynamic Time Warping algorithm adapted
# from: https://github.com/talcs/simpledtw (MIT License)
######################################################################


def warp(fixation_XY, text_block):
    """

    Map fixations to word centers using [Dynamic Time
    Warping](https://en.wikipedia.org/wiki/Dynamic_time_warping). This finds a
    monotonically increasing mapping between fixations and words with the
    shortest overall distance, effectively resulting in *m* subsequences.
    Fixations are then assigned to the lines that their mapped words belong
    to, effectively assigning subsequences to text lines in chronological
    order.

    """
    word_XY = np.array(text_block.word_centers, dtype=int)
    n1 = len(fixation_XY)
    n2 = len(word_XY)
    cost = np.zeros((n1 + 1, n2 + 1))
    cost[0, :] = np.inf
    cost[:, 0] = np.inf
    cost[0, 0] = 0
    for fixation_i in range(n1):
        for word_i in range(n2):
            distance = np.sqrt(sum((fixation_XY[fixation_i] - word_XY[word_i]) ** 2))
            cost[fixation_i + 1, word_i + 1] = distance + min(
                cost[fixation_i, word_i + 1],
                cost[fixation_i + 1, word_i],
                cost[fixation_i, word_i],
            )
    cost = cost[1:, 1:]
    warping_path = [[] for _ in range(n1)]
    while fixation_i > 0 or word_i > 0:
        warping_path[fixation_i].append(word_i)
        possible_moves = [np.inf, np.inf, np.inf]
        if fixation_i > 0 and word_i > 0:
            possible_moves[0] = cost[fixation_i - 1, word_i - 1]
        if fixation_i > 0:
            possible_moves[1] = cost[fixation_i - 1, word_i]
        if word_i > 0:
            possible_moves[2] = cost[fixation_i, word_i - 1]
        best_move = np.argmin(possible_moves)
        if best_move == 0:
            fixation_i -= 1
            word_i -= 1
        elif best_move == 1:
            fixation_i -= 1
        else:
            word_i -= 1
    warping_path[0].append(0)
    for fixation_i, words_mapped_to_fixation_i in enumerate(warping_path):
        candidate_Y = list(word_XY[words_mapped_to_fixation_i, 1])
        fixation_XY[fixation_i, 1] = max(set(candidate_Y), key=candidate_Y.count)
    return fixation_XY[:, 1]


methods = {
    "chain": chain,
    "cluster": cluster,
    "merge": merge,
    "regress": regress,
    "segment": segment,
    "split": split,
    "stretch": stretch,
    "warp": warp,
}



import setuptools

with open("README.md", encoding="utf-8") as file:
    long_description = file.read()

setuptools.setup(
    name="eyekit",
    use_scm_version={
        "write_to": "eyekit/_version.py",
        "write_to_template": '__version__ = "{version}"',
        "fallback_version": "???",
    },
    author="Jon Carr",
    author_email="jcarr@sissa.it",
    description="A Python package for analyzing reading behavior using eyetracking data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jwcarr.github.io/eyekit/",
    license="MIT",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3.6",
    install_requires=["cairocffi>=1.1", "numpy>=1.19"],
    setup_requires=["setuptools_scm"],
)
